// 2004 - Chaim Gingold

#include "Plant.h"

#include "fColor.h"
#include <assert.h>
#include "chaim.h"

#include "PlantSeg.h"

#include "Miter.h"

Plant::Plant( PlantWorld * w, fPoint2 base ) : world(w), base(base)
{
	sim = w->sim ;
	nNodes = 0 ;
}

Plant::~Plant()
{
	for( int i=0; i<nNodes; i++ )
	{
		Rigid_Body			* rb = nodes[i].rb ;
		Pt2Pt_Constraint	* parentPt2Pt = nodes[i].parentPt2Pt ;
		Angle_Spring_Force	* angleSpring = nodes[i].angleSpring ;

		if (rb)
		{
			sim->remove(*rb) ;
			delete rb ;
			rb = 0 ;
		}

		if (parentPt2Pt)
		{
			sim->remove(*parentPt2Pt) ;
			delete parentPt2Pt ;
			parentPt2Pt = 0 ;
		}

		if (angleSpring)
		{
			sim->remove(*angleSpring) ;
			delete angleSpring ;
			angleSpring = 0 ;
		}
	}
}

void
Plant::GrowPlant ( int forLevel )
{
	float gutter = pix(8) ;

	// make root
	int i = AddNode( -1, base, kBasePlantRadius, kPlantBaseMass ) ; //fPoint2(b.GetCenterX(),b.bottom), 32 ) ;
	
	fruitDeadZone.Set( base ) ;
	fruitDeadZone.Inset( -kBasePlantRadius*3, -kBasePlantRadius*3 ) ;

	if ( i != -1 )
	{
		nodes[i].rb -> inv_m = 0 ;
		nodes[i].rb -> inv_I = 0 ;
	}

	switch( forLevel % kNumLevels )
	{
	case 0:
		GrowAllTips( 3, gutter ) ;
		ForkAllTips( 2, DegToRad(60), gutter ) ;
		GrowAllTips( 3, gutter ) ;
		GrowAllTips( 3, gutter ) ;
		break ;

	case 1:
		ForkAllTips( 3, DegToRad(60), gutter ) ;

		GrowAllTips( 7, gutter ) ;

		break ;

	case 2:
		GrowAllTips( 3, gutter ) ;
		ForkAllTips( 4, DegToRad(180), gutter ) ;
		GrowAllTips( 3, gutter ) ;
		//ForkAllTips( 2, DegToRad(90), gutter ) ;
		GrowAllTips( 3, gutter ) ;
		break ;

	case 3:
		ForkAllTips( 2, DegToRad(60), gutter ) ;
		GrowAllTips( 2, gutter ) ;
		ForkAllTips( 2, DegToRad(60), gutter ) ;
		GrowAllTips( 2, gutter ) ;
		ForkAllTips( 2, DegToRad(60), gutter ) ;
		GrowAllTips( 4, gutter ) ;
		break ;

	case 4:
		GrowAllTips( 4, gutter ) ;
		ForkAllTips( 3, DegToRad(180), gutter ) ;
		GrowAllTips( 3, gutter ) ;
		ForkAllTips( 2, DegToRad(90), gutter ) ;
		GrowAllTips( 3, gutter ) ;
		break ;

	case 5:
		GrowAllTips( 2, gutter ) ;
		ForkAllTips( 2, DegToRad(60), gutter ) ;
		GrowAllTips( 2, gutter ) ;
		ForkAllTips( 2, DegToRad(60), gutter ) ;
		GrowAllTips( 2, gutter ) ;
		ForkAllTips( 2, DegToRad(60), gutter ) ;
		GrowAllTips( 3, gutter ) ;
		break ;

	default:
		assert(0) ;
		break ;

	}

	//
	MakeTipsHoles() ;

	ComputeWaypoints() ;

	// mark leafs
	for( int i=0; i<nNodes; i++ )
	{
		nodes[i].isLeaf = ! GetHasChild(i) ;
	}
}

void
Plant::MakeTipsHoles ( void )
{
	for( int i=0; i<nNodes; i++ )
	{
		if ( GetHasChild(i)==0 ) nodes[i].hole = 1 ;
	}
}

int
Plant::CountFruitSlots ( int * numFull )
{
	int e=0, f=0 ;

	for( int i=0; i<nNodes; i++ )
	{
		int h = nodes[i].hole ;

		if (h==1) e++ ;
		else if (h==2) f++ ;
	}

	if (numFull) *numFull=f ;

	return e ;
}

bool
Plant::FitFruit ( fPoint2 p )
{
	for( int i=0; i<nNodes; i++ )
	{
		if ( nodes[i].hole == 1 && nodes[i].GetLoc().Dist(p) < kFitFruitDelta )
		{
			nodes[i].hole = 2 ;
			return true ;
		}
	}

	return false ;
}

void
Plant::GrowFruitAtAllTips ( void )
{
	for( int i=0; i<nNodes; i++ )
	{
		if ( GetHasChild(i)==0 ) MakeFruit(i) ;
	}
}

void
Plant::GrowAllTips ( int grow, float gutter )
{
	// uninited, too.
	SetColors(0) ;

	// mark tips to grow
	for( int i=0; i<nNodes; i++ )
	{
		nodes[i].color = GetHasChild(i) ? 0 : 1 ;
	}

	// grow them
	for( int i=0; i<nNodes; i++ )
	{
		if ( nodes[i].color ) GrowTip( i, grow, gutter ) ;
	}
}

void 
Plant::ForkAllTips ( int nchildren, fRad spread, float gutter )
{
	// uninited, too.
	SetColors(0) ;

	// mark tips to grow
	for( int i=0; i<nNodes; i++ )
	{
		nodes[i].color = GetHasChild(i) ? 0 : 1 ;
	}

	// grow them
	for( int i=0; i<nNodes; i++ )
	{
		if ( nodes[i].color ) ForkTip( i, nchildren, spread, gutter ) ;
	}
}

void
Plant::ForkTip ( int from, int nchildren, fRad spread, float gutter )
{
	assert( from < nNodes && from >= 0 ) ;

	// figure out angles
	fRad starta, stepa ;

	fRad origina = -DegToRad(90) ;
	
	int pi = nodes[from].parent ;

	if ( pi != -1 ) origina = nodes[pi].GetLoc().Angle( nodes[from].GetLoc() ) ;

	starta = origina - spread/2 ;
	stepa  = spread / (float) (nchildren-1) ;

	//
	fPoint2 myLoc = nodes[from].GetLoc() ;
	float   r  = nodes[from].r * kPlantRadiusScalePerStep ;
	float dist = nodes[from].r + r + gutter ;
	float	m  = nodes[from].mass * kMassPlantScalePerStep ;

	for( int i=0; i<nchildren; i++ )
	{
		fPoint2 p ;
		p.SetPolar( myLoc, starta + stepa * (float)i, dist ) ;

		AddNode ( from, p, r, m ) ;
	}
}

int
Plant::GrowTip ( int from, int grow, float gutter )
{
	assert( from < nNodes && from >= 0 ) ;

	fRad a = DegToRad(-90) ;

	float mass = nodes[from].mass ;

	int last = from ;

	// find angle
	int p = nodes[from].parent ;

	if ( p != -1 ) a = nodes[p].GetLoc().Angle( nodes[from].GetLoc() ) ;

	//
	while ( grow-- )
	{
		mass *= kMassPlantScalePerStep ;

		float   r = nodes[from].r ;
		fPoint2 c ;

		c.SetPolar( nodes[from].GetLoc(), a, nodes[from].r + r + gutter ) ;

		from = AddNode( from, c, r * kPlantRadiusScalePerStep, mass ) ;

		if ( from == -1 ) return last ;

		last = from ;
	}

	return last ;
}

int
Plant::AddNode ( int parent, fPoint2 loc, float r, float mass )
{
	if ( nNodes==kMaxNodes ) return -1 ;

	int i = nNodes++ ;

	nodes[i].parent = parent ;
	nodes[i].rb = NewRigidCircle( loc, r*.5, mass ) ;
	nodes[i].r = r ;
	nodes[i].createdLoc = loc ;
	nodes[i].mass = mass ;


	if ( nodes[i].rb )
	{
		nodes[i].rb -> m_Pbody -> m_collision_plane = kPlantCollisionMask ;

		sim -> add( *nodes[i].rb ) ;
	}

	if ( parent != -1 )
	{
		fPoint2 pin ;
		pin.SetToBetween( nodes[i].GetLoc(), nodes[parent].GetLoc(), .5 ) ;

		nodes[i].parentPt2Pt
			= new Pt2Pt_Constraint(
				Handle( *nodes[i].rb,	   nodes[i].rb->world_space_to_body_space( f2v(pin) ) ) ,
				Handle( *nodes[parent].rb, nodes[parent].rb->world_space_to_body_space( f2v(pin) ) ) ) ;

		if (nodes[i].parentPt2Pt) sim->add( *nodes[i].parentPt2Pt ) ;

		
		nodes[i].angleSpring
			= new Angle_Spring_Force(
				Handle( *nodes[i].rb,	   Vec2(0,0) ) ,
				Handle( *nodes[parent].rb, Vec2(0,0) ) ,
				kAngleSpringConst ,
				kAngleSpringDamp ) ;

		if (nodes[i].angleSpring) sim->add( *nodes[i].angleSpring ) ;
	}

	// wire in spring constraint
	return i ;
}

void
Plant::MakeFruit ( int atNode )
{
	world->AddObj( new Fruit( world, GetNodeLoc(atNode) ) ) ;
}

void
Plant::SkinPlant ( void )
{
	for( int i=0; i<nNodes; i++ )
	{
		int p = nodes[i].parent ;

		fPoint2 il = nodes[i].GetLoc() ;

		fRad a ;

		if ( p == -1 )
		{
			a=kPI ;
		}
		else
		{
			fPoint2 pl = nodes[p].GetLoc() ;

			a = GetLastMiterAngle( pl, il ) ;
		}

		//
		fPoint2 m, n ;

		m.SetPolar( il, a,     nodes[i].r/2 ) ;
		n.SetPolar( il, a+kPI, nodes[i].r/2 ) ;

		nodes[i].leftMeat  = m ;
		nodes[i].rightMeat = n ;
	}
}

void
Plant::OutlineFlesh ( void )
{
	fColor::SetToBlend( fColor::green(), fColor::black(), .75 ) ;

	for( int i=0; i<nNodes; i++ )
	{
		int p = nodes[i].parent ;

		if ( p != -1 )
		{
			Line( nodes[i].leftMeat,  nodes[p].leftMeat  ) ;
			Line( nodes[i].rightMeat, nodes[p].rightMeat ) ;
			Line( nodes[i].GetLoc(),  nodes[p].GetLoc()  ) ;
		}
	}
}

void
Plant::RenderFlesh ( void )
{
	fColor lGreen, rGreen, cGreen ;

	lGreen.SetToBetween( fColor::green(), fColor::black(), .35 ) ;
	rGreen.SetToBetween( fColor::green(), fColor::black(), .25 ) ;
	cGreen.SetToBetween( fColor::green(), fColor::black(), .75 ) ;


	for( int i=0; i<nNodes; i++ )
	{
		int p = nodes[i].parent ;

		if ( p != -1 )
		{
			BeginPolygon() ;
			lGreen.SetColor() ;
			Vertex( nodes[i].leftMeat  ) ;
			Vertex( nodes[p].leftMeat  ) ;
			Vertex( nodes[p].GetLoc()  ) ;
			Vertex( nodes[i].GetLoc()  ) ;
			EndPolygon() ;


			BeginPolygon() ;
			rGreen.SetColor() ;
			Vertex( nodes[i].GetLoc()  ) ;
			Vertex( nodes[p].GetLoc()  ) ;
			Vertex( nodes[p].rightMeat ) ;
			Vertex( nodes[i].rightMeat ) ;
			EndPolygon() ;

			cGreen.SetColor() ;
			Line( nodes[i].leftMeat,  nodes[p].leftMeat  ) ;
			Line( nodes[i].rightMeat, nodes[p].rightMeat ) ;
			Line( nodes[i].GetLoc(),  nodes[p].GetLoc()  ) ;
		}
	}	
}

void
Plant::Render ( void )
{
	SkinPlant() ;
	RenderFlesh() ;
	//OutlineFlesh() ;

	bool wireframe = false ;

	for( int i=0; i<nNodes; i++ )
	{
		fPoint2 p = nodes[i].GetLoc() ;

		if (wireframe)
		{
			PenWidth(3) ;
			fColor::black() -> SetColor() ;
			FrameCircle( p, nodes[i].r ) ;

			int parent = nodes[i].parent ;

			if ( parent != -1 ) Line( p, nodes[parent].GetLoc() ) ;
			PenWidth(1) ;
		}

		//
		int h = nodes[i].hole ;

		if ( h > 0 )
		{
			if ( h==1 )
			{
				fColor::black() -> SetColor() ;
				FrameCircle( p, kFruitRadius+pix(3) ) ;
				FrameCircle( p, kFruitRadius+pix(6) ) ;
			}
			else
			{
				DrawSun( p, kFruitRadius + pix(512.0f)*gFadeLevel ) ;

				/*
				//fColor::SetToBlend( fColor::orange(), fColor::white(), .5 ) ;
				fColor::yellow()->SetColor() ;
				PaintCircle( p, kFruitRadius ) ;
				fColor::purple()->SetColor() ;
				FrameCircle( p, kFruitRadius ) ;

				SetOpacity(.25) ;
				PenWidth(3) ;
				fColor::yellow() -> SetColor() ;
				FrameCircle( p, kFruitRadius+pix(9) ) ;
				SetOpacity(1) ;*/
			}
		}
	}

	PenWidth(1) ;
}

bool
Plant::GetHasChild ( int n ) const
{
	for( int i=0; i<nNodes; i++ )
	{
		if ( nodes[i].parent == n ) return true ;
	}

	return false ;
}

void
Plant::SetColors ( int c )
{
	for( int i=0; i<kMaxNodes; i++ )
	{
		nodes[i].color = c ;
	}
}

bool
Plant::PinToPlant ( fPoint2 * loc, float * maxd, bool onlyAdjEdges, int oldEdge[2], float oldFudge ) const
{
	fPoint2 p = *loc ;

	fPoint2 best ;
	bool found = false ;

	int oldNodeA = -1 ;
	int oldNodeB = -1 ;

	if (oldEdge)
	{
		oldNodeA = oldEdge[0] ;
		oldNodeB = oldEdge[1] ;
	}

	for( int i=0; i<nNodes; i++ )
	{
		if ( nodes[i].parent == -1 ) continue ;

		int ai = i ,
			bi = nodes[i].parent ;

		fPoint2 a = nodes[ai].GetLoc() ,
				b = nodes[bi].GetLoc() ;

		float d = p.DistToLineSegment( a, b ) ;

		if (onlyAdjEdges)
		{
			if ( ( ai==oldNodeA || bi==oldNodeB ) ||
				 ( bi==oldNodeA || ai==oldNodeB ) || oldNodeA==-1 || oldNodeB==-1
				 )
			{
				//d -= oldFudge ;
				int x = 1 ;
			}
			else
			{
				//d += oldFudge ;
				//d = *maxd + 1 ;
				continue ;
			}
		}
		else
		{
			if ( ( ai==oldNodeA || bi==oldNodeB ) ||
				 ( bi==oldNodeA || ai==oldNodeB ) )
			{
				d += oldFudge ;
			}
			else
			{
				d -= oldFudge ;
			}
		}

		if ( d < *maxd )
		{
			*maxd = d ;
			found = true ;

			best = p ;
			best.SnapToLineSegment( a, b ) ;

			if (oldEdge)
			{
				oldEdge[0] = ai ;
				oldEdge[1] = bi ;
			}
		}
	}

	if ( found ) *loc = best ;

	return found ;
}

bool
Plant::PinToPlant ( fPoint2 * loc, float * maxd ) const
{
	fPoint2 p = *loc ;

	fPoint2 best ;
	bool found = false ;

	for( int i=0; i<nNodes; i++ )
	{
		if ( nodes[i].parent == -1 ) continue ;

		fPoint2 a = nodes[i].GetLoc() ,
				b = nodes[ nodes[i].parent ].GetLoc() ;

		float d = p.DistToLineSegment( a, b ) ;

		if ( d < *maxd )
		{
			*maxd = d ;
			found = true ;

			best = p ;
			best.SnapToLineSegment( a, b ) ;
		}

		if ( nodes[i].isLeaf )
		{
			d = nodes[i].GetLoc().Dist(p) - kTipExtraStickyness  ;
			if (d<0) d=0 ;

			if ( d < *maxd )
			{
				*maxd = d ;
				found = true ;

				best = nodes[i].GetLoc() ;
			}
		}
	}

	if ( found ) *loc = best ;

	return found ;
}

Rigid_Body*
Plant::FindClosestPlantBody ( fPoint2 loc, float * d, int plantEdge[2] )
{
	if ( nNodes==0 ) return 0 ;

	float bestd ;
	Rigid_Body * res = 0 ;

	for( int i=0; i<nNodes; i++ )
	{
		if ( plantEdge )
		{
			int pe0 = plantEdge[0], pe1 = plantEdge[1] ;

			if ( i != pe0 && i != pe1 ) continue ;
		}

		float d = nodes[i].GetLoc().Dist(loc) ;

		if ( !res || d < bestd )
		{
			bestd = d ;
			res = nodes[i].rb ;
		}
	}

	if (res) *d = bestd ;

	return res ;
}

int
Plant::GetClosestNode  ( fPoint2 loc ) const
{
	if ( nNodes==0 ) return -1 ;

	float bestd = nodes[0].GetLoc().Dist(loc) ;
	int b=0 ;

	for( int i=1; i<nNodes; i++ )
	{
		float d = nodes[i].GetLoc().Dist(loc) ;

		if ( d < bestd )
		{
			bestd = d ;
			b=i ;
		}
	}

	return b ;
}

int
Plant::FindWaypointNode ( int from, int to ) const
{
	assert( from < nNodes && from >= 0 ) ;
	assert( to < nNodes && to >= 0 ) ;

	return nodes[from].waypoints[to] ;
}

fPoint2
Plant::GetNodeLoc ( int node ) const
{
	assert( node < nNodes && node >= 0 ) ;

	return nodes[node].GetLoc() ;
}

void
Plant::ComputeWaypoints ( void )
{
	int c[ kMaxNodes ] ; // color

	// for every destination node
	for( int dst=0; dst<nNodes; dst++ )
	{
		// identity
		nodes[dst].waypoints[dst] = dst ;

		// seed flood fill
		for( int i=0; i<nNodes; i++ ) c[i] = -1 ;
		c[dst] = 0 ; // seed it

		// do a flood fill coloring
		bool touched ;

		do
		{
			touched = false ;

			for( int a=0; a<nNodes; a++ )
			for( int b=0; b<nNodes; b++ )
			{
				// arc: a --> b

				if ( c[a] != -1 && c[b] == -1 && // correct color state
					 ( nodes[a].parent==b || nodes[b].parent==a ) // and connected
				    )
				{
					// bleed the color
					c[b] = c[a] + 1 ;

					// adjacency: *dst* -- (some stuff) -- *a* -- *b*
					// so: to get to *dst* from *b*, you need to go to *a*

					// note it in the waypoint list
					nodes[b].waypoints[dst] = a ;

					// keep doing fill
					touched = true ;
				}
			}

		} while ( touched ) ;
	}
}