#ifndef __PlantH__
#define __PlantH__

// 2004 - Chaim Gingold

#include "SimHelp.h"
#include "chaim.h"

const int kPlantCollisionMask = 0 ;
const int kObjCollisionMask   = 2 ;

// plant
const float kTipExtraStickyness = pix(8) ;

const float kPlantBaseMass = 1 ;

const float kMassPlantScalePerStep   = .5 ;
const float kPlantRadiusScalePerStep = .75 ;

const float kBasePlantRadius = pix(32) ;

// plant angle springs
#if 1 // two players
const float kAngleSpringConst = 1000.0f ; // 1000f
const float kAngleSpringDamp  = 50.0f ; //50.0f ;
#else // one player
const float kAngleSpringConst = 100.0f ; // 100f
const float kAngleSpringDamp  = 25.0f ; //5.0f ;
#endif

// fruit goal fitting
const float kFitFruitDelta = pix(4) ; // 1/3 of fruit radius


class PlantWorld ;


	/*
	 *	Plant
	 */

class Plant
{

public:
	Plant( PlantWorld *, fPoint2 base ) ;
	~Plant() ;

	void GrowPlant ( int forLevel ) ;

	Rigid_Body* FindClosestPlantBody ( fPoint2 loc, float * d, int plantEdge[2] ) ; // d is output only
	// if plantEdge is specified, will only check against these bodies

	int		GetClosestNode  ( fPoint2 loc ) const ;
	int		FindWaypointNode( int from, int to ) const ;
	fPoint2	GetNodeLoc		( int node ) const ;

	void GrowAllTips ( int grow, float gutter ) ;
	void ForkAllTips ( int nchildren, fRad spread, float gutter ) ;
	void GrowFruitAtAllTips ( void ) ;

	void ForkTip ( int from, int nchildren, fRad spread, float gutter ) ;
	int  GrowTip ( int from, int grow, float gutter ) ;
	// from what node, grow how many new nodes,
	// returns index of new tip

	int AddNode ( int parent, fPoint2 loc, float r, float mass ) ; // returns index, -1 for fail

	void Render ( void ) ;

	bool PinToPlant ( fPoint2 * loc, float * maxd ) const ;

	bool PinToPlant ( fPoint2 * loc, float * maxd, bool onlyAdjEdges, int oldEdge[2], float oldFudge ) const ;
	// writes edges in

	fPoint2 GetRootLoc ( void ) const { return nodes[0].GetLoc() ; }

	void MakeFruit ( int atNode ) ;

	void MakeTipsHoles ( void ) ;

	int CountFruitSlots ( int * numFull=0 ) ; // returns num empty

	bool FitFruit ( fPoint2 ) ;

	fRect fruitDeadZone ; // where fruit won't attach automatically

	void SkinPlant ( void ) ;
	void RenderFlesh ( void ) ;
	void OutlineFlesh ( void ) ;

private:

	bool GetHasChild ( int ) const ;

	enum
	{
		kMaxNodes = 128
	} ;

	PlantWorld * world ;
	Sim * sim ;

	int nNodes ;

	fPoint2 base ;


	void SetColors ( int ) ;

	struct node
	{
		node()
		{
			rb=0 ;
			parentPt2Pt = 0 ;
			angleSpring = 0 ;
			color = 0 ;

			hole = 0 ;

			for ( int i=0; i<kMaxNodes; i++ ) waypoints[i] = -1 ;

			isLeaf=false ;
		}

		bool isLeaf ;

		int hole ; // 0: no, 1: yes&empty, 2: yes&full

		int color ;

		Rigid_Body			* rb ;
		Pt2Pt_Constraint	* parentPt2Pt ;
		Angle_Spring_Force	* angleSpring ;

		float mass ;
		int   parent ;
		float r ;

		fPoint2 leftMeat, rightMeat ;

		fPoint2 GetLoc ( void ) const { return v2f(rb->x) ; } //return bs2ws( rb, fPoint2(0,0) ) ; }
		// where it is now

		fPoint2 createdLoc ; // where it was made

		int waypoints[ kMaxNodes ] ;
		// if you want to go from here to any other node,
		// look up your destination in the table,
		// and you get an index back for the waypoint.
		// i.e. waypoint = waypoints[dest]

	} nodes[ kMaxNodes ] ;

	void ComputeWaypoints ( void ) ;

} ;


#endif