// 2004 - Chaim Gingold

#include "SimHelp.h"

#include "editor.h"
#include "tesselator.h"

fPoint2 bs2ws ( Rigid_Body * rb, fPoint2 bs )
{
	Handle h( *rb, f2v(bs) ) ;
	return v2f( h.x_worldS() ) ;
}

fPoint2 ws2bs ( Rigid_Body * rb, fPoint2 ws )
{
	return v2f( rb->world_space_to_body_space( f2v(ws) ) ) ;
}

fPoint2 v2f ( Vec2 v )    { return fPoint2(v.x(),v.y()) ; }
Vec2    f2v ( fPoint2 f ) { return Vec2(f.x,f.y) ; }

Rigid_Body* NewRigidPoly ( fPoint2 * v, int n, float density )
{
	tesselation		   * tess ;
	tesselation_result * tessRes ;

	tess = BeginTesselation() ;
	BeginContour( tess ) ;
	
	for ( int i=0; i<n; i++ ) Vertex2f( tess, v[i].x, v[i].y ) ;

	EndContour( tess ) ;
	tessRes = EndTesselation( tess ) ;

	Rigid_Body * rb = makeConcaveRigidBody( tessRes, density ) ;

	FreeTesselationResult( tessRes ) ;

	return rb ;
}

Rigid_Body* NewRigidBox ( fRect r )
{
	return new Rigid_Body(
		make_Body_Box( Vec2( r.right, r.bottom ), Vec2( r.left, r.top ) ) ) ;
//		Vec2(-77.5f,0.0f), 0.0f,0.0f,0.0f ) ;
}

Rigid_Body* NewRigidCircle ( fPoint2 c, float r, float mass )
{
	Rigid_Body * rb = new Rigid_Body( make_Body_Sphere( r, Vec2(0,0) ) ) ;

	rb->inv_m = 1/mass ;
	rb->x = f2v(c) ;

	return rb ;
}


	/*
	 *
	 */

static void glRigidBody ( Rigid_Body * ) ;

static void glBody1v ( Rigid_Body *rb, Body *body ) ;


void FrameRigidBody ( Rigid_Body * b )
{
	//glBegin( GL_LINE_LOOP) ;

	glBody1v( b, b->m_Pbody.get() ) ;

	//glEnd() ;
}

void PaintRigidBody ( Rigid_Body * b )
{
	//glBegin( GL_POLYGON ) ;

	glBody1v( b, b->m_Pbody.get() ) ;
	
	//glEnd() ;
}

void
glBody1v(Rigid_Body *rb, Body *body)
{
    if(body)
    {
		{
			Body_Hull *p = dynamic_cast<Body_Hull*>(body);
			if(p)
			{
				glBegin(GL_TRIANGLE_FAN);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Box *p = dynamic_cast<Body_Box*>(body);
			if(p)
			{
				glBegin(GL_TRIANGLE_FAN);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Sphere *p = dynamic_cast<Body_Sphere*>(body);
			if(p)
			{
				glBegin(GL_TRIANGLE_FAN);
				int total_steps = 16;
				Vec2 center = p->x();
				glVertex2f(center.x(), center.y());
				for(int a = 0;a<total_steps;++a)
				{
					float Angle = 2.0f * 3.14159265358979323846f * ((float)a / (float)(total_steps - 1));
					float Radius = p->r();
					Vec2 arm(Radius*cos(Angle), Radius*sin(Angle));
					arm += center;
					glVertex2f(arm.x(), arm.y());
				}
				glEnd();
			}
		}

        Body_List *l = dynamic_cast<Body_List*>(body);
        if(l)
        {
            {for(int Index = 0;
                 Index < l->num_bodies();
                 ++Index)
            {
                glBody1v(rb, &l->body(Index));
            }}
        }
    }
}