
	// debug physics sim
	if (0)
	{
		GL::gl_error_check();
		AJB_CONSOLE("num forces " << (int)gWorld->sim->m_binary_forces.size() << endl);
		static int frame_count = 0;
		AJB_CONSOLE("frame count " << frame_count++ << endl);
		gWorld->sim->draw(GL::GL_Physics2d_Draw( kWidth, f2v(GetBounds().GetCenter()) ));	//GL_Physics2d_Draw is a callback object
		GL::gl_error_check();
	}

	// render pathfinding info
	if (0)
	{
		fPoint2 from = GetMouseDownLoc() ;
		fPoint2 to   = GetMouseLoc() ;

		/*
		int fromi = gWorld->plant->GetClosestNode( from ) ;
		int toi   = gWorld->plant->GetClosestNode( to   ) ;

		int midi  = gWorld->plant->FindWaypointNode( fromi, toi ) ;
		
		if ( midi != -1 )*/ // -- just tests plants graph pathfinding.
		{
			//fPoint2 mid = gWorld->plant->GetNodeLoc(midi) ;
			
			fPoint2 mid = gWorld->GetWaypoint( from, to, pix(16) ) ;

			fColor::red()->SetColor() ;
			PaintCircle( from, pix(3) ) ;
			Line( from, mid ) ;

			fColor::green()->SetColor() ;
			PaintCircle( mid, pix(3) ) ;
			Line( mid, to ) ;

			fColor::blue()->SetColor() ;
			PaintCircle( to, pix(3) ) ;
		}
	}

	// select stuff with recct
	if ( 0 && GetIsMouseDown() )
	{
		fRect q( GetMouseDownLoc().x, GetMouseDownLoc().y, GetMouseLoc().x, GetMouseLoc().y ) ;

		q.Sort() ;

		fColor::blue()->SetColor() ;
		q.Frame() ;

		gWorld->UnselectAll( 1 ) ;
		gWorld->Select( 1, q ) ;
		gWorld->Select( 1, GetMouseDownLoc() ) ;
	}


	/*
	 *	PlantSeg
	 */

class PlantSeg
{

public:
	PlantSeg( PlantWorld*, fPoint2 base ) ; // does a fork
	PlantSeg( PlantWorld*, fPoint2 base, int depthLeft ) ; // does a linear list

	void LockInPlace ( void ) ;

	void Render ( void ) const ;

	PlantSeg* GetLeaf ( void ) { if ( child[0] ) return child[0]->GetLeaf() ; return this ; }

	Rigid_Body* GetRigidBody ( void ) const { return rb ; }

	fPoint2 GetBase   ( void ) const ; // in world space
	fPoint2 GetBaseBS ( void ) const { return constructedWithBase ; } // in body space

	PlantSeg* FindSegByRigidBody ( Rigid_Body * ) ;

	fPoint2 GetBaseForChild ( int c ) { return childBase[c] ; }

	void AddChild( int c, PlantSeg * ps ) ;

	int CountChildren ( void ) const ;

private:
	Rigid_Body * rb ;

	enum
	{
		kMaxChildren = 2
	} ;

	fPoint2	childBase[ kMaxChildren ] ;

	fPoint2 constructedWithBase ;
		// base of where it was created; use GetBase
		// to find transformed base

	void MakeStandSpring ( void ) ;

	Spring_Force		* standSpring ;

	Pt2Pt_Constraint	* childSprings[ kMaxChildren ] ;
	PlantSeg			* child[ kMaxChildren ] ;

	PlantWorld* world ;

} ;



	/*
	 *	PlantSeg
	 */

PlantSeg::PlantSeg( PlantWorld* world, fPoint2 base, int depthLeft )
	: constructedWithBase(base) ,
	  world(world)
{
	float gutter = 1 ; // distance to next segment

	for( int i=0; i<kMaxChildren; i++ )
	{
		child[i] = 0 ;
		childSprings[i] = 0 ;
	}

	int spawnChildren = depthLeft > 0 ? 1 : 0 ;

	// parameterize segment
	float h = 64 ;
	float w = 32 ;

	// construct the segment outline

	const int kMaxV = 8 ;
	fPoint2 rp[ kMaxV ] ;
	int n = 0 ;


	{
		float topInset=14, botInset=16 ;

		n = 6 ;

		rp[0] = fPoint2( 0, 0 ) ; // base
		rp[1] = fPoint2( -w/2, -botInset ) ; // left edge
		rp[2] = fPoint2( -w/2, -topInset - h ) ;
		rp[3] = fPoint2(    0, -h ) ; // top tip
		rp[4] = fPoint2(  w/2, -topInset - h ) ; // right edge
		rp[5] = fPoint2(  w/2, -botInset ) ;

		childBase[0] = fPoint2( 0, -h - gutter ) ;
	}

	// offset rp to base (we assume base is origin)
	fPoint2::OffsetArray( rp, n, base ) ;
	fPoint2::OffsetArray( childBase, kMaxChildren, base ) ;

	// make the rigid body
	rb = NewRigidPoly( rp, n, kPlantDensity ) ;

/*
	// simple box segments
	fRect r( base.x-w/2, base.y-h, base.x+w/2, base.y ) ;
	rb = NewRigidBox( r ) ;

	springRestLen = gutter ;

	childBase[0] = fPoint2( base.x, base.y-h - gutter ) ;
*/

	if (rb)
	{
		world -> sim -> add(*rb) ;

		rb -> inv_m = 2.0f ;
	}

	// make children
	if ( spawnChildren > 0 )
	{
		for( int i=0; i<spawnChildren; i++ )
		{
			AddChild( 0, new PlantSeg( world, childBase[0], depthLeft-1 ) ) ;
		}
	}

	MakeStandSpring() ;
}

PlantSeg::PlantSeg( PlantWorld* world, fPoint2 base )
	: constructedWithBase(base) ,
	  world(world)
{
	float gutter = 1 ; // distance to next segment

	for( int i=0; i<kMaxChildren; i++ )
	{
		child[i] = 0 ;
		childSprings[i] = 0 ;
	}

	// parameterize segment
	float h = 64 ;
	float w = 32 ;

	// construct the segment outline

	const int kMaxV = 8 ;
	fPoint2 rp[ kMaxV ] ;
	int n = 0 ;



	{
		float topInset=20, botInset=48 ;

		n = 8 ;

		w *= 3 ;

		rp[0] = fPoint2( 0, 0 ) ; // base
		rp[1] = fPoint2( -w/2, -botInset ) ; // left edge
		rp[2] = fPoint2( -w/2, -topInset - h ) ;
		
		rp[3] = fPoint2( -w/4, -h ) ; // top inset
		rp[4] = fPoint2(    0, -topInset - h ) ; // top tip
		rp[5] = fPoint2(  w/4, -h ) ; // top inset

		rp[6] = fPoint2(  w/2, -topInset - h ) ; // right edge
		rp[7] = fPoint2(  w/2, -botInset ) ;

		childBase[0] = fPoint2( -w/4, -h - gutter ) ;
		childBase[1] = fPoint2(  w/4, -h - gutter ) ;
	}

	// offset rp to base (we assume base is origin)
	fPoint2::OffsetArray( rp, n, base ) ;
	fPoint2::OffsetArray( childBase, kMaxChildren, base ) ;

	// make the rigid body
	rb = NewRigidPoly( rp, n, kPlantDensity ) ;



	if (rb)
	{
		world -> sim -> add(*rb) ;

		rb -> inv_m = 2.0f ;
	}

	MakeStandSpring() ;
}

void
PlantSeg::MakeStandSpring ( void )
{
	if ( CountChildren() > 0 ) return ;
//	assert( !standSpring ) ;

	standSpring = new Spring_Force(
			Handle( *world->earth, f2v( GetBase() ) ) ,
			Handle( *rb, f2v(constructedWithBase) ) ,
			kPlantStandSpringK ,
			kPlantStandSpringDamp ,
			kPlantStandSpringRestLen
			) ;

	world->sim->add( *standSpring ) ;
}

void
PlantSeg::AddChild ( int i, PlantSeg * ps )
{
	child[i] = ps ;

	if ( child[i] && child[i]->rb )
	{
		fPoint2 b = child[i]->constructedWithBase ;

		childSprings[i] = new Pt2Pt_Constraint(
			Handle( *rb, f2v(b) ) ,
			Handle( *(child[i]->rb), f2v(b) ) ) ;

		if ( childSprings[i] ) world->sim->add( *childSprings[i] );
	}
}

int PlantSeg::CountChildren ( void ) const
{
	int c=0 ;
	for( int i=0; i<kMaxChildren; i++ ) if ( child[i] ) c++ ;
	return c ;
}

void PlantSeg::LockInPlace ( void ) // just this segment -- irreversible
{
	if (rb)
	{
		rb -> inv_m = 0 ;
		rb -> inv_I = 0 ;
	}
}

void PlantSeg::Render ( void ) const
{
	fColor::SetToBlend( fColor::black(), fColor::green(), .5 ) ;

	if (rb) FrameRigidBody(rb) ;

	for( int i=0; i<kMaxChildren; i++ ) if ( child[i] ) child[i] -> Render() ;

	//
	fColor::red()->SetColor() ;

	PaintCircle( GetBase(), 5 ) ;
}

fPoint2
PlantSeg::GetBase ( void ) const
{
	return bs2ws( rb, constructedWithBase ) ;	
}

PlantSeg*
PlantSeg::FindSegByRigidBody ( Rigid_Body * r )
{
	if ( r == rb ) return this ;

	for( int i=0; i<kMaxChildren; i++ )
	{
		if ( !child[i] ) continue ; 

		PlantSeg * s = child[i] -> FindSegByRigidBody (r) ;

		if (s) return s ;
	}

	return 0 ;
}